/*
 * kgraft_patch_bsc1044878
 *
 * Fix for bsc#1044878
 *
 *  Upstream commit:
 *  - none: upstream, i.e. commit
 *    2851940ffee313e0ff12540a8e11a8c54dea9c65 ("netfilter: allow logging from
 *    non-init namespaces") is not affected.
 *
 *  SLE12-SP2 commit:
 *  - none: SLE12-SP2 also isn't affected as it's based on upstream.
 *
 *  SLE12-SP1 commit:
 *  3d586f4339f640b76430b2c8df148e2cf73647d1
 *
 *  SLE12-LTSS commit:
 *  0ebe3fc994b7ddb62e267a0a3e9af446e6bd6bc5
 *
 *  Copyright (c) 2017 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/sysctl.h>
#include <net/netfilter/nf_log.h>
#include <net/net_namespace.h>

#if !IS_MODULE(CONFIG_NETFILTER_XT_TARGET_LOG)
#error "KGR patch supports only CONFIG_NETFILTER_XT_TARGET_LOG=m."
#elif !IS_ENABLED(CONFIG_SYSCTL)
#error "KGR patch is pointless for !CONFIG_SYSCTL."
#endif

#define KGR_PATCHED_MODULE "xt_LOG"

static struct ctl_table_header **kgr_nf_log_sysctl_fhdr;
static struct nf_logger *kgr_ipt_log_logger;
static struct nf_logger *kgr_ip6t_log_logger;

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "xt_LOG:nf_log_sysctl_fhdr", (void *)&kgr_nf_log_sysctl_fhdr },
	{ "xt_LOG:ipt_log_logger", (void *)&kgr_ipt_log_logger },
#if IS_ENABLED(CONFIG_IP6_NF_IPTABLES)
	{ "xt_LOG:ip6t_log_logger", (void *)&kgr_ip6t_log_logger },
#endif
};

void kgr_log_net_exit(struct net *net)
{
#ifdef CONFIG_SYSCTL
	/*
	 * Fix bsc#1044878
	 *  -1 line, +1 line
	 */
	if (net_eq(net, &init_net) && *kgr_nf_log_sysctl_fhdr)
		unregister_sysctl_table(*kgr_nf_log_sysctl_fhdr);
#endif
	nf_log_unset(net, kgr_ipt_log_logger);
#if IS_ENABLED(CONFIG_IP6_NF_IPTABLES)
	nf_log_unset(net, kgr_ip6t_log_logger);
#endif
}

static int kgr_patch_bsc1044878_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1044878_module_notify(struct notifier_block *nb,
	unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1044878_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1044878_module_nb = {
	.notifier_call = kgr_patch_bsc1044878_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1044878_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1044878_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1044878_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1044878_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1044878_module_nb);
}
